' RNM.BAS  Version 6D
'
' Generates random numbers, with inputted number of outcomes (1/P),
'  number of trails per run or set, number of runs or sets, and with
'  or without replacement.  Uses random number routine RANDNUM, described
'  in the routine.  Several possible output formats.
'
' Modified from RNM.FTN written initially by Jim Kennedy during the 1970's
'  while at FRNM.
'
' Version V6A   4/5/91  written in QuickBasic by JK.
' Version V6B   4/7/91  changed output format to allow 4 digits of
'                       compressed output.
' Version V6C   5/30/93 changed randomization without replacement to be one
'                       pass randomly selecting from remaining numbers --
'                       was two passes of random shuffling before.
' Version V6D   3/30/05 added cr to lf
'
'
' ************************ PROGRAM NOTES ****************************
'
' On input, default values (i.e., from just hitting enter) are in [].
'
' If 0 or a negative no. is entered at the input lines, then the
' program exits.

' Maximum no. trials or events per run or set is 200 in this version.
'
' Normal output gives info. on input parameters, Randnum seed, etc.  Output
' can be to specified file or device.  If file exits, then output is
' appended.
'
' Output formats depend on no. digits of possible outcomes.  In normal
' output, one or two digits for no. possible outcomes (i.e, less than 100)
' has 20 columns per line and all other range of possible outcomes have 10
' columns per output line -- maximum of 5 digits for no. possible outcomes.
'
' Compressed format has no descriptors or titles and is for input into other
' computer programs.  Option for compressed output is offered only if more
' than 25 runs or sets of numbers.  Compressed output can be for 1 to 4
' digit nos. and fills 60 output columns (unless the no. trials per run or
' set times the no. of digits in the possible outcomes is not an even
' multiple of 60).
'
' If random nos. are without replacement and fewer nos. requested than
' possible outcomes, then the program asks if the randomly selected nos.
' should be randomly ordered.  This is for the case of selecting some
' cases from a large no. of cases -- the important information is which
' cases  were selected, not the order.
'
' If random nos. are without replacement and the no. of trials per run or set
' is not an even multiple of the no. of possible outcomes, then the extra or
' odd trials are randomly filled with the outcomes without replacement --
' so there will be no more than one extra of each possible outcome.
'
' Strategy for random nos. without replacement is to fill an array with the
' nos. and then randomly swap nos. (unless, as noted above, the data looks
' like random selection of a sample from a large population -- in which case
' the option of not random swapping or ordering is offered to give a nice
' sequential list).
'
' The Randnum function is documented in the subroutine -- seems like a very
' good random no. generator.  Subroutine Randset used to initialize Randnum.
' Initial setting comes from clock and QuickBasic RND function -- seems ok.
' RND gives very different nos. just a few seconds apart.
'
' The reference for the algorithm for randomly selecting n trials from a
' larger no. N trials is given in the program below when used.
'
' ************************* VARIABLES *******************************
'
'  rnum(55) rsuf(54) iran jran ntran& -- used in Randnum function. (shared)
'                   See documentation in the function.
'  nresult(200)  -- array for all random nos. in a run or set.
'  ntrials       -- no. random nos. in a run or set.
'  itrials       -- index for counting to ntrials.
'  np            -- no. possible outcomes (i.e., 1/P).
'  nsets         -- no. runs or sets of ntrials.
'  isets         -- index for counting to nsets.
'  tseed         -- random no. seed -- obtained from QuickBasic TIMER
'                   function (which is no. seconds from midnight) and is
'                   used to randomize the Quick Basic RND function which is
'                   used to fill the random tables for the Randnum function.
'  nadd          -- no. odd or extra trials when random nos. are without
'                   replacement and the no. of trials per run (ntrials) is
'                   not an integer multiple of the no. of possible outcomes
'                   (np).
'  rn rr p x     -- used in the algorithm for selecting a sample from a
'                   larger group without replacement -- used to get
'                   extra trials when nadd is not zero.
'  ii            -- index counter used for counting to get nadd extra trials
'                   if needed, and later for indexing which output format
'                   array is to be used.
'  afmt(6) afmt2 afmt3 -- output format arrays for various output.
'  ncol          -- tells how many colums of output there are -- set based
'                   on no. of digits and whether or not output is compressed.
'  areplace      -- "Y" or "N" for random nos. with or without replacement.
'  aorder        -- "Y" or "N" for randomly swapping nos. without
'                   replacement when the no. nos. is smaller than the no.
'                   of possible outcomes -- i.e., drawing random sample.
'  acompress     -- "Y" or "N" for making compressed output for file for
'                   input into some computer program.
'
' **************************** PROGRAM ********************************

  CONST aver$ = "6D"                        'Set version number
  DECLARE FUNCTION Randnum! ()              'Random no. Generator
  DECLARE SUB Randset ()                    'Initializes random no generator
 
'**** Set variable characteristics
 
  OPTION BASE 1:  DEFINT I-N:  DEFSTR A:
  DIM SHARED rnum(55) AS LONG, rshuf(0 TO 54) AS LONG   'For RANDNUM
  DIM SHARED iran, jran, ntran&                         'For RANDNUM
  DIM nresult(200)
  DIM afmt(6) AS STRING * 6
 
'**** Clear screen and get input parameters

  CLS
  PRINT "RNM("; aver; ")>No. Possible Outcomes (1/P) (0=Stop)>"; : INPUT "", np
    IF np <= 0 THEN PRINT "RNM> Finished": END
 
  INPUT "RNM>Trials per Run or Set>", ntrials
    IF ntrials <= 0 THEN END
    IF ntrials > 200 THEN PRINT "MAX No. Trials per Set is 200": END
 
  INPUT "RNM>With Replacement (Y or N) [Y]>", areplace
    IF areplace = "" THEN areplace = "Y" ELSE areplace = UCASE$(areplace)
    IF areplace <> "Y" AND areplace <> "N" THEN END
 
  IF areplace = "N" AND ntrials < np THEN      'may not want reordering here
    INPUT "RNM>Order in Set Randomized (Y or N) [N]>", aorder
    IF aorder = "" THEN aorder = "N" ELSE aorder = UCASE$(aorder)
    IF aorder <> "Y" AND aorder <> "N" THEN END
  END IF
 
  INPUT "RNM>Number of Runs or Sets>", nsets
    IF nsets <= 0 THEN END
    IF nsets > 25 THEN                   'Format for use in computer programs
      INPUT "RNM>Compressed Output (Y or N [N])>", acompress
      IF acompress = "" THEN acompress = "N" ELSE acompress = UCASE$(acompress)
    END IF
 
  INPUT "RNM>Output to [con]>", aoutfile
    IF aoutfile = "" THEN aoutfile = "CON" ELSE aoutfile = UCASE$(aoutfile)
    OPEN aoutfile FOR APPEND AS #2

' *** Get randnum seed and initialize randnum
    tseed = TIMER: RANDOMIZE (tseed):  Randset
   
'**** Write conditions to output file -- if not compressed output

  IF acompress <> "Y" THEN
   PRINT #2, CHR$(13); CHR$(10); CHR$(13); CHR$(10); CHR$(13); CHR$(10) 'Line feeds for top of page
   PRINT #2, "        ****************** RNM Version ("; aver; ") ********************"
   PRINT #2, "                    Run at "; TIME$; " on  "; DATE$
   PRINT #2,
   PRINT #2, "          No. Outcomes (1/P) ="; np; "         Replacement = "; areplace
   PRINT #2, "          No. Trials per Set ="; ntrials; "        No. Sets    ="; nsets
   PRINT #2, "          Randnum Seed       ="; tseed; "  Output to:  "; aoutfile
   PRINT #2, CHR$(13); CHR$(10)
  END IF

'**** Generate random numbers

  FOR isets = 1 TO nsets               'Do each set or run

' *** Get random nos. with replacement
    IF areplace = "Y" THEN
      FOR itrial = 1 TO ntrials
        nresult(itrial) = 1 + FIX(Randnum * np)
      NEXT itrial
   
' *** Get random nos. without replacement
    ELSE
      IF ntrials >= np THEN                 'Handle sampling case below
        FOR itrial = 1 TO ntrials           'List outcomes to be re-ordered
          nresult(itrial) = (itrial MOD np) + 1
        NEXT itrial
      END IF
     
      nadd = ntrials MOD np              'nadd is odd trials relative to np
      IF nadd > 0 THEN
     
'  ** Handle odd trials.  Randomly select nadd nos. from np nos. where
'     nadd < np.  Algorithm from Bissel, 1986, Appl. Stat. vol 35 pp.73-75
        rn = np + 1!:    rr = np - nadd + 1!
        ii = 0                           'ii is no. nos. selected so far
        DO
          x = Randnum:  p = 1
          DO
            rr = rr - 1!:   rn = rn - 1!
            p = p * (rr / rn)
          LOOP UNTIL (p <= x)
          ii = ii + 1
          nresult(ntrials - ii + 1) = np + 1 - FIX(rn)
          rr = rr + 1!
        LOOP UNTIL (ii = nadd)
      END IF   ' end nadd loop
      
'  **  Randomize order for nos. without replacement -- swap remaining nos.
      IF ntrials >= np OR aorder = "Y" THEN
        FOR itrial = 1 TO ntrials - 1
          SWAP nresult(itrial), nresult(FIX((Randnum * (ntrials + 1 - itrial)) + itrial))
        NEXT itrial
      END IF  ' end ntrials
    END IF  ' end get random nos.
 
'**** Print output

' *** Set format arrays
    afmt(1) = " ##"                     ' Format for 2 digit nos
    afmt(2) = " #####"                  ' Format for 3 or more digits
    afmt(3) = "#"                       ' Compressed format for 1 digit
    afmt(4) = "##"                      ' Compressed format for 2 digits
    afmt(5) = "###"                     ' Compressed format for 3 digits
    afmt(6) = "####"                    ' Compressed format for 4 digits
    afmt2 = "####: ## - "               ' For new set of nos.
    afmt3 = "     ### - "               ' For second and later lines per set

' *** set no. output columns (ncol) and ii, which indexes output format array
    IF acompress <> "Y" THEN            ' Normal output
      IF np < 100 THEN ncol = 20: ii = 1 ELSE ncol = 10: ii = 2
      IF ntrials < ncol THEN ncol = ntrials     ' do not print extra columns
    ELSE                                ' Compressed output
      ncol = FIX(LOG(np) / LOG(10#)) + 1:  ' ncol is no. digits (no log10)
      ii = 2 + ncol                        ' set for output formal array
      ncol = 60 / ncol
    END IF

' *** print column header nos.
    IF isets = 1 AND acompress <> "Y" THEN
      PRINT #2, "           ";
      FOR itrial = 1 TO ncol
        PRINT #2, USING RTRIM$(afmt(ii)); itrial;
      NEXT itrial
      PRINT #2, : PRINT #2, "           ";
      FOR itrial = 1 TO ncol
        IF ii = 1 THEN PRINT #2, " --";  ELSE PRINT #2, "   ---";
      NEXT itrial
    END IF
'
' *** print random nos. -- keep adding to the line until get no. columns
    IF acompress <> "Y" THEN PRINT #2, CHR$(13); CHR$(10): PRINT #2, USING afmt2; isets; 1;
    FOR itrial = 1 TO ntrials
      IF ((itrial - 1) MOD ncol = 0) AND itrial <> 1 THEN   'mod for new line
        PRINT #2,
        IF acompress <> "Y" THEN PRINT #2, : PRINT #2, USING afmt3; itrial;
      END IF
      PRINT #2, USING RTRIM$(afmt(ii)); nresult(itrial);
    NEXT itrial
    PRINT #2,
 
  NEXT isets   ' end main loop for each run or set

'**** Send notice to terminal
 
  LOCATE 24
  PRINT :
  PRINT "RNM>Completed"

END

DEFDBL D
'  FUNCTION RANDNUM
'
'  Generates random nos. uniformly distributed between 0 and 1.
'  Uses "subtractive algorithm" (p. 171) and "Randomization by Shuffling --
'  Algorithm B" (p. 32) in "The Art of Computer Programming," 1981 by
'  Donald Knuth.  Got some of combined algorithm from "MicroSoft QuickBasic
'  Programmer's Toolbox," 1988, by John Craig (pp. 355-364).
'
'  Function needs to be setup with subroutine RANDSET.
'  Program statements that need to be in the main program are given in
'  RANDSET.
'
'  Passed freqency, runs up and down, and collision randomness tests as per
'  Knuth.
'
'  To get random integer values use IX = INT(RANDNUM * N) + 1
'
'  Do not use mod function to get integers as that (i.e., lower bits)
'  did not pass randomness tests.
'
'  Program is significantly faster if incorporated into main program rather
'  than called as function.
'
'  Version 1A,  January 12, 1989 -- Jim Kennedy
'
'  Variables:
'
'  NTRAN&       Long integer holding random integer between 0 and 1 billion.
'               Shared.  Divided by 1E-09 to get real no.
'  RNUM(55)     Long integer array to hold the various random integers that
'               are subtracted to get a new random integer.  Moved through
'               in a cycle, subtracting relative nos. 55 and 24 (i.e., 31
'               apart.  Shared.
'  IRAN and JRAN  integer indexes for RNUM -- kept 31 nos. apart.  Shared.
'  RSHUF(0 to 54) Long integer array to hold nos. from RNUM that are
'               shuffled.  Each no. from RNUM is shuffled with a no.
'               in RSHUF.  Shared.
'  IRANS        Random integer between 0 and 54 obtained from last random
'               no. (i.e., from NTRAN&) used as index for shuffling.
'               Mod function used to generate IRANS.  This is not totally
'               random by itself , but is much faster than multiplication
'               and combined with subtraction method gives good randomness.
'
'  Programming Notes:
'
'  Did some experiments to find fastest running time.  Eg. multiplication
'  faster than division.   This routine is much faster than some other
'  generators I tried.  Could be made slightly faster by using nos. 1 and 60
'  instead of 55 and 24 -- as per Knuth p. 28.  But apparently 24 and 55
'  are more widely used and have been tested more.
'
FUNCTION Randnum
  
     irans = (ntran& MOD 55)
     ntran& = rnum(iran) - rnum(jran)
     IF ntran& < 0 THEN ntran& = ntran& + 1000000000
     rnum(iran) = ntran&
     IF iran = 1 THEN iran = 55 ELSE iran = iran - 1
     IF jran = 1 THEN jran = 55 ELSE jran = jran - 1
     SWAP ntran&, rshuf(irans)
     Randnum = ntran& * 1E-09

END FUNCTION

' SUBROUTINE RANDSET
'
' Initializes arrays for RANDNUM random no. generator function.
'
' Variables defined in RANDNUM function.
' Uses QuickBasic RND function to get initial nos. for the arrays.
'
' Need to get seed and randomize for starting.  Relevant lines for
' main program are given below as comments.
'
' This program is taken from "The Art of Computer Programming," 1981, by
' Donald Knuth and "MicroSoft QuickBasic Programmer's Tools," 1988 by
' John Craig.  More explanation in the RANDNUM function.
'
' Version 1A -- January 12, 1989 -- Jim Kennedy
'
' Version 1B -- July 13, 1989.  Fixed minor bug that was not initializing
'               NTRAN& randomly (had mispelled it).
'
'  The following lines need to be in the main program.
'
'OPTION BASE 1
' DECLARE FUNCTION RANDNUM! ()         'Random no. Generator
' DECLARE SUB RANDSET ()               'Initialized random no generator
'
' DIM SHARED RNUM(55) AS LONG, RSHUF(0 TO 54) AS LONG         ' for RANDNUM
' DIM SHARED IRAN, JRAN, NTRAN&                               ' for RANDNUM
'
' INPUT "PROGNAME>ENTER RANDNUM SEED>", ISEED             'Seed for RANDSET
' RANDOMIZE ISEED         
' CALL RANDSET                                ' Initialize RANDNUM function
'
'
SUB Randset

  FOR i = 1 TO 55
    rnum(i) = 1E+09 * RND                   'Initialize no. array
    rshuf(i - 1) = 1E+09 * RND              'Initialize shuffle array
  NEXT i
  iran = 55: jran = 24:  ntran& = 1E+09 * RND   ' to do first shuffle
  FOR i = 1 TO 550                          'Warm up generator
    R = Randnum
  NEXT i

END SUB

